@@ -30,3 +30,12 @@ class URLRequired(RequestException):
 
 class TooManyRedirects(RequestException):
     """Too many redirects."""
+
+class MissingSchema(RequestException, ValueError):
+    """The URL schema (e.g. http or https) is missing."""
+
+class InvalidSchema(RequestException, ValueError):
+    """See defaults.py for valid schemas."""
+    
+class InvalidURL(RequestException, ValueError):
+    """ The URL provided was somehow invalid. """
\ No newline at end of file
